/**
 *	@file camera.h Camera: definio de cmeras e clculos relacionados.
 *
 *	@author
 *			- Maira Noronha
 *			- Thiago Bastos
 *			- Mauricio Carneiro
 *
 *	@date
 *			Criado em:			30 de Novembro de 2002
 *			ltima Modificao:	22 de Janeiro de 2003
 *
 *	@version 2.0
 */

#ifndef _CAMERA_H_
#define _CAMERA_H_

#include "algebra.h"


/************************************************************************/
/* Tipos Exportados                                                     */
/************************************************************************/
/**
 *   Camera no mesmo formato do OpenGL.
 */
struct _Camera
{
	/* Definio da cmera */

    /**
     * Posio da cmera (do observador).
     */
	Vector eye;
	/**
	 * Direo para onde o observador est olhando.
	 */
	Vector at;
	/**
	 * Direo perpendicular a at que indica a orientao vertical da cmera.
	 */
	Vector up;
	/**
	 * Abertura da cmera (ngulo de viso), de 0 a 180.
	 */
	double fovy;
	/**
	 * Distncia de eye at nearp (onde a cena ser projetada).
	 */
	double nearp;
	/**
	 * Distncia de eye at farp (background).
	 */
	double farp;
	/**
	 * Largura da tela em pixels.
	 */
	double screenWidth;
	/**
	 * Altura da tela em pixels
	 */
	double screenHeight;

	/* Estado interno */

	/**
	 * Direo x
	 */
	Vector xAxis;
	/**
	 * Direo y
	 */
	Vector yAxis;
	/**
	 * Direo z
	 */
	Vector zAxis;
	/**
	 * Origem do near plane.
	 */
	Vector nearOrigin;
	/**
	 * Vetor u do near plane.
	 */
	Vector nearU;
	/**
	 * Vetor v do near plane.
	 */
	Vector nearV;
	/**
	 * Origem do far plane.
	 */
	Vector farOrigin;
	/**
	 * Vetor normal ao far plane.
	 */
	Vector farNormal;
	/**
	 * Vetor u do far plane.
	 */
	Vector farU;
	/**
	 * Vetor v do far plane.
	 */
	Vector farV;
};

typedef struct _Camera * Camera;


/************************************************************************/
/* Funes Exportadas                                                   */
/************************************************************************/
/**
 *	Cria uma nova cmera usando as propriedades especificadas.
 *
 *	@param eye Posio da cmera (do observador).
 *	@param at Direo para onde o observador est olhando.
 *	@param up Direo perpendicular a at que indica a orientao vertical da cmera.
 *	@param fovy Abertura da cmera (ngulo de viso), de 0 a 180.
 *	@param nearp Distncia de eye at nearp.
 *	@param farp Distncia de eye at farp.
 *	@param screenWidth Largura da tela em pixels.
 *	@param screenHeight Altura da tela em pixels.
 *
 *	@return Handle da cmera criada.
 */
Camera camCreate( Vector eye, Vector at, Vector up, double fovy, double nearp, double farp,
					int screenWidth, int screenHeight );

/**
 *	Obtm informaes sobre o far plane definido para uma cmera.
 *
 *	@param origin [out]Retorna a origem do far plane.
 *	@param normal [out]Retorna um vetor normal ao far plane.
 *	@param u [out]Retorna o vetor u do far plane.
 *	@param v [out]Retorna o vetor v do far plane.
 */
void camGetFarPlane( Camera camera, Vector *origin, Vector *normal, Vector *u, Vector *v );

/**
 *	Obtm a posio de uma cmera.
 *
 *	@return Posio do 'eye' da cmera.
 */
Vector camGetEye( Camera camera );

/**
 *	Obtm um raio saindo do eye de uma cmera e passando por um pixel especificado.
 *
 *	@param x Posio X do pixel por onde o raio deve passar.
 *	@param y Posio Y do pixel por onde o raio deve passar.
 *
 *	@return Um raio saindo de eye e passando pelo pixel (x,y) no near plane.
 */
Vector camGetRay( Camera camera, double x, double y );

/**
 *	Obtm a largura da tela de uma cmera, em pixels.
 */
int camGetScreenWidth( Camera camera );

/**
 *	Obtm a altura da tela de uma cmera, em pixels.
 */
int camGetScreenHeight( Camera camera );

/**
 *	Destri uma cmera criada com camCreate().
 */
void camDestroy( Camera camera );

void camGetOGLPerspective(Camera camera, double* fovy, double* aspect, double* zNear, double* zFar);

void camGetOGLLookAt(Camera camera, double* eyex, double* eyey,double* eyez, double* centerx,double* centery,double* centerz, double* upx,double* upy,double* upz);

#endif

