/**
 *	@file material.h Material: manuteno de materiais.
 *
 *	@author
 *			- Maira Noronha
 *			- Thiago Bastos
 *			- Mauricio Carneiro
 *
 *	@date
 *			Criado em:			1 de Dezembro de 2002
 *			ltima Modificao:	4 de Junho de 2003
 *
 *	@version 2.0
 */

#ifndef _MATERIAL_H_
#define _MATERIAL_H_

#include "color.h"
#include "image.h"
#include "algebra.h"
#include "mainaux.h"

/************************************************************************/
/* Tipos Exportados                                                     */
/************************************************************************/
/**
 *   Material com o qual  feito um objeto.
 */
struct _Material
{
	/**
     *  Textura do material.
     */
	Image texture;
    /**
    *  Textura do material para Zbuffer.
    */
	TextureImage* texturaZbuffer;

	/**
     *  Textura de normais.
     */
	Image normal;

	/**
     *  Cor base do material (difusa).
     */
	Color diffuseColor;
	/**
     *  Cor do brilho especular do material.
     */
	Color specularColor;

	/**
     *  Coeficiente que define o brilho especular.
     */
	double specularExponent;
	/**
     *  Fator de refletividade do material.
     */
	double reflectionFactor;
	/**
     *  ndice de refrao do material.
     */
	double refractionFactor;
	/**
     *  Opacidade do material.
     */
	double opacityFactor;
};

typedef struct _Material * Material;


/************************************************************************/
/* Funes Exportadas                                                   */
/************************************************************************/
/**
 *	Cria um novo material com as propriedades especificadas.
 *
 *	@param texture Imagem contendo textura do material (pode ser NULL).
 *	@param diffusecolor Cor base do material ( substituido pela textura, quando presente).
 *	@param specularColor Cor do brilho especular para este material.
 *	@param specularExponent Coeficiente que define o brilho especular.
 *	@param reflectionFactor Fator de refletividade do material: 0 se no reflete nada,
 *				1 se totalmente reflexivo).
 *	@param refractionFactor ndice de refrao do material, para materiais transparentes.
 *	@param opacityFactor Opacidade do material (1 para 100% opaco, 0 para 100% transparente).
 *
 *	@return Handle para o material criado.
 */
Material matCreate( Image texture, TextureImage* texturaZbuffer,Color diffuseColor, 
					Color specularColor, double specularExponent,
					double reflectionFactor, double refractionFactor, double opacityFactor );

/**
 *	Obtm a cor difusa para um objeto composto de um material.
 *
 *	@param material Handle para o material do objeto.
 *	@param textureCoordinate Coordenada de textura calculada para o objeto em questo.
 *
 *	@return Cor difusa do objeto num certo ponto.
 */
Color matGetDiffuse( Material material, Vector textureCoordinate );

/**
 *	Obtm a cor do brilho especular de um material.
 */
Color matGetSpecular( Material material );

/**
 *	Obtm o coeficiente do brilho especular (o expoente N).
 */
double matGetSpecularExponent( Material material );

/**
 *	Obtm o fator de refletividade do material: 1.0f para 100% reflexivo.
 */
double matGetReflectionFactor( Material material );

/**
 *	Obtm o ndice de refrao de um material (usado para implementar transparncia).
 */
double matGetRefractionIndex( Material material );

/**
 *	Obtm a opacidade de um material: 1.0f para 100% opaco, 0.0f para 100% transparente.
 */
double matGetOpacity( Material material );

/**
 *	Destri um material criado com matCreate().
 */
void matDestroy( Material material );

int temTextura(Material material);

#endif
