/**
 *	@file object.c Object: definio e operaes com primitivas.
 *		As primitivas suportadas atualmente so: esferas, tringulos e paraleleppedos.
 *
 *	@author
 *			- Maira Noronha
 *			- Thiago Bastos
 *			- Mauricio Carneiro
 *
 *	@date
 *			Criado em:			01 de Dezembro de 2002
 *			ltima Modificao:	22 de Janeiro de 2003
 *
 *	@version 2.0
 */

#include "object.h"
//#include "algebra.h"

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <gl/glut.h>                 /* prototypes do OpenGL */
#include "mainaux.h"



/************************************************************************/
/* Constantes Privadas                                                  */
/************************************************************************/
#define MIN( a, b ) ( ( a < b ) ? a : b )
#define MAX( a, b ) ( ( a > b ) ? a : b )

#ifndef EPSILON
#define EPSILON	1.0e-3
#endif

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif




enum
{
	TYPE_UNKNOWN,
	TYPE_SPHERE,
	TYPE_TRIANGLE,
	TYPE_BOX
};




/************************************************************************/
/* Definio das Funes Exportadas                                     */
/************************************************************************/
Object objCreateSphere( int material, const Vector center, double radius, int bumpFlag, int blendFlag )
{
	Object object;
	Sphere *sphere;

	object = (struct _Object *)malloc( sizeof(struct _Object) );
	sphere = (Sphere *)malloc( sizeof(Sphere) );

	sphere->center = center;
	sphere->radius = radius;

	object->type = TYPE_SPHERE;
	object->bump = bumpFlag;
	object->blend = blendFlag;
	object->material = material;
	object->data = sphere;

	return object;
}


Object objCreateTriangle( int material, const Vector v0, const Vector v1, const Vector v2, 
						                const Vector tex0, const Vector tex1, const Vector tex2, int bumpFlag , int blendFlag )
{
	Object object;
	Triangle *triangle;

	object = (struct _Object *)malloc( sizeof(struct _Object) );
	triangle = (Triangle *)malloc( sizeof(Triangle) );

	triangle->v0 = v0;
	triangle->v1 = v1;
	triangle->v2 = v2;

	triangle->tex0 = tex0;
	triangle->tex1 = tex1;
	triangle->tex2 = tex2;

	object->type = TYPE_TRIANGLE;
	object->bump = bumpFlag;
	object->blend = blendFlag;
	object->material = material;
	object->data = triangle;

	return object;
}


Object objCreateBox( int material, const Vector bottomLeft, const Vector topRight, int bumpFlag , int blendFlag)
{
	Object object;
	Box *box;

	object = (struct _Object *)malloc( sizeof(struct _Object) );
	box = (Box *)malloc( sizeof(Box) );

	box->bottomLeft = bottomLeft;
	box->topRight = topRight;

	object->type = TYPE_BOX;
	object->material = material;
	object->data = box;
	object->bump = bumpFlag;
	object->blend = blendFlag;

	return object;
}


double objIntercept( Object object, Vector eye, Vector ray )
{
	switch( object->type )
	{
	case TYPE_SPHERE:
		{
			Sphere *s = (Sphere *)object->data;

			double a, b, c, delta;
			double distance = -1.0;

			Vector fromSphereToEye;

			fromSphereToEye = algSub( eye, s->center );

			a = algDot( ray, ray );
			b = ( 2.0 * algDot( ray, fromSphereToEye ) );
			c = ( algDot( fromSphereToEye, fromSphereToEye ) - ( s->radius * s->radius ) );

			delta = ( ( b * b ) - ( 4 * a * c ) );

			if( fabs( delta ) <= EPSILON )
			{
				distance = ( -b / (2 * a ) );
			}
			else if( delta > EPSILON )
			{
				double root = sqrt( delta );
				distance = MIN( ( ( -b + root ) / ( 2 * a ) ), ( ( -b - root ) / ( 2.0 * a ) )  );
			}

			return distance;
		}

	case TYPE_TRIANGLE:
		{
			Triangle *t = (Triangle *)object->data;
			
			double dividend, divisor;
			double distance = -1.0;

			Vector v0ToV1 = algSub( t->v1, t->v0 );
			Vector v1ToV2 = algSub( t->v2, t->v1 );
			Vector normal = algCross( v0ToV1, v1ToV2 );
			Vector eyeToV0 = algSub( t->v0, eye );

			dividend = algDot( eyeToV0, normal );
			divisor = algDot( ray, normal );

			if( divisor <= -EPSILON )
			{
				distance = ( dividend / divisor );
			}

			if( distance >= 1.0 )
			{
				double a0, a1, a2;

				Vector v2ToV0 = algSub( t->v0, t->v2 );
				Vector p = algAdd( eye, algScale( distance, ray ) );
				Vector n0 = algCross( v0ToV1, algSub( p, t->v0 ) );
				Vector n1 = algCross( v1ToV2, algSub( p, t->v1 ) );
				Vector n2 = algCross( v2ToV0, algSub( p, t->v2 ) );

				normal = algUnit(normal);
				a0 = ( 0.5 * algDot( normal, n0 ) );
				a1 = ( 0.5 * algDot( normal, n1 ) );
				a2 = ( 0.5 * algDot( normal, n2 ) );

				if ( (a0>0) && (a1>0) && (a2>0) ) 
					return distance;
				else
					return -1.0;
			}

			return distance;
		}

	case TYPE_BOX:
		{
			Box *box = (Box *)object->data;

			double xmin = box->bottomLeft.x;
		    double ymin = box->bottomLeft.y;
		    double zmin = box->bottomLeft.z;
		    double xmax = box->topRight.x;
		    double ymax = box->topRight.y;
		    double zmax = box->topRight.z;
			
			double x, y, z;
			double distance = -1.0;

			if( ray.x > EPSILON || -ray.x > EPSILON )
			{
				if( ray.x > 0 )
				{
					x = xmin;
					distance = ( ( xmin - eye.x ) / ray.x );
				}
				else
				{
					x = xmax;
					distance = ( ( xmax - eye.x ) / ray.x );
				}

				if( distance > EPSILON )
				{
					y = ( eye.y + ( distance * ray.y ) ); 
					z = ( eye.z + ( distance * ray.z ) ); 
                    if( ( y >= ymin ) && ( y <= ymax ) && ( z >= zmin ) && ( z <= zmax ) )
                        return distance;
				}
			}

			if( ray.y > EPSILON || -ray.y > EPSILON )
			{
				if( ray.y > 0 )
				{
					y = ymin;
					distance = ( ( ymin - eye.y ) / ray.y );
				}
				else
				{
					y = ymax;
					distance = ( ( ymax - eye.y ) / ray.y );
				}
				
				if( distance > EPSILON )
				{
					x = ( eye.x + ( distance * ray.x ) ); 
					z = ( eye.z + ( distance * ray.z ) ); 
				    if( ( x >= xmin ) && ( x <= xmax ) && ( z >= zmin ) && ( z <= zmax ) )
					    return distance;
				}

			}

			if( ray.z > EPSILON || -ray.z > EPSILON )
			{
				if( ray.z > 0 )
				{
					z = zmin;
					distance = ( (zmin - eye.z ) / ray.z );
				}
				else
				{
					z = zmax;
					distance = ( ( zmax - eye.z ) / ray.z );
				}

				if( distance > EPSILON )
				{
					x = ( eye.x + ( distance * ray.x ) ); 
					y = ( eye.y + ( distance * ray.y ) ); 
				    if( ( x >= xmin ) && ( x <= xmax ) && ( y >= ymin ) && ( y <= ymax ) )	
						return distance;
				}
			}

			return -1.0;
		}
	
	default:
		/* Tipo de Objeto Invlido: nunca deve acontecer */
		return -1.0;
	}
}


Vector objNormalAt( Object object, Vector point )
{
	if( object->type == TYPE_SPHERE )
	{
		Sphere *sphere = (Sphere *)object->data;

		return algScale( ( 1.0 / sphere->radius ),
					algSub( point, sphere->center ) );
	}
	else if ( object->type == TYPE_TRIANGLE )
	{
		Triangle *triangle = (Triangle *)object->data;

		return algCross( algSub( triangle->v1, triangle->v0 ),
					algSub( triangle->v2, triangle->v0 ) );
	}
	else if ( object->type == TYPE_BOX )
	{
		Box *box = (Box *)object->data;
		/* Seleciona a face mais prxima de point */
		if( fabs( point.x - box->bottomLeft.x ) < EPSILON )
		{
			return algVector( -1, 0, 0, 1  );
		}
		else if( fabs( point.x - box->topRight.x ) < EPSILON )
		{
			return algVector( 1, 0, 0, 1 );
		}
		else if( fabs( point.y - box->bottomLeft.y ) < EPSILON )
		{
			return algVector( 0, -1, 0, 1 );
		}
		else if( fabs( point.y - box->topRight.y ) < EPSILON )
		{
			return algVector( 0, 1, 0, 1 );
		}
		else if( fabs( point.z - box->bottomLeft.z ) < EPSILON )
		{
			return algVector( 0, 0, -1, 1 );
		}
		else if( fabs( point.z - box->topRight.z ) < EPSILON )
		{
			return algVector( 0, 0, 1, 1 );
		}
		else
		{
			return algVector( 0, 0, 0, 1 );
		}
	} 
	else
	{
		/* Tipo de Objeto Invlido: nunca deve acontecer */
		return algVector( 0, 0, 0, 1 );
	}
}

Vector objTextureCoordinateAt( Object object, Vector point )
{
	double a = 0.0;
	double b = 0.0;
	double c = 0.0;


	if( object->type == TYPE_SPHERE )
	{
		Vector normal = objNormalAt( object, point );
		double phi = atan2( normal.y, normal.x );
		double theta = atan2( sqrt( ( normal.x * normal.x ) + ( normal.y * normal.y ) ), normal.z );

		return algVector( ( 3 * ( 1 + phi / M_PI ) ), ( 3 * ( 1 - theta / M_PI ) ), 0, 1 );
	} 
	else if( object->type == TYPE_TRIANGLE )
	{
		Triangle *t = (Triangle *)object->data;
		Vector v0ToV1 = algSub( t->v1, t->v0 );
		Vector v1ToV2 = algSub( t->v2, t->v1 );
		Vector normal = algCross( v0ToV1, v1ToV2 );
		Vector p = point ;


		double a0, a1, a2 ,at;

		Vector v2ToV0 = algSub( t->v0, t->v2 );

		Vector n0 = algCross( v1ToV2, algSub( p, t->v1 ) );
		Vector n1 = algCross( v2ToV0, algSub( p, t->v2 ) );
		Vector n2 = algCross( v0ToV1, algSub( p, t->v0 ) );

		normal = algUnit(normal);
		a0 = ( 0.5 * algDot( normal, n0 ) );
		a1 = ( 0.5 * algDot( normal, n1 ) );
		a2 = ( 0.5 * algDot( normal, n2 ) );
		at = a0 + a1 + a2 ;
		a0 /= at;
		a1 /= at;
		a2 /= at;

		/* Encontrando p - referente a textura */

		v0ToV1 = algSub( t->tex1, t->tex0 );
		v1ToV2 = algSub( t->tex2, t->tex1 );
		v2ToV0 = algSub( t->tex0, t->tex2 );
		n0 = algScale(a0 , t->tex0);
		n1 = algScale(a1 , t->tex1);
		n2 = algScale(a2 , t->tex2);

		p = algAdd(n0,algAdd(n1,n2));

		return p;

	} 
	else if( object->type == TYPE_BOX )
	{
		Box *box = (Box *)object->data;

		double xmin = box->bottomLeft.x;
		double ymin = box->bottomLeft.y;
		double zmin = box->bottomLeft.z;
		double xmax = box->topRight.x;
		double ymax = box->topRight.y;
		double zmax = box->topRight.z;

		if( ( fabs( point.x - xmin ) < EPSILON ) || ( fabs( point.x - xmax ) < EPSILON ) )
		{
			return algVector( ( ( point.y - ymin ) / ( ymax - ymin ) ), ( ( point.z - zmin ) / ( zmax - zmin ) ), 0, 1 );
		}
		else if( ( fabs( point.y - ymin ) < EPSILON ) || ( fabs( point.y - ymax ) < EPSILON ) )
		{
			return algVector( ( ( point.z - zmin ) / ( zmax - zmin ) ), ( ( point.x - xmin ) / ( xmax - xmin ) ), 0, 1 );
		}
		else if( ( fabs( point.z - zmin ) < EPSILON ) || ( fabs( point.z - zmax ) < EPSILON ) )
		{
			return algVector( ( ( point.x - xmin ) / ( xmax - xmin ) ), ( ( point.y - ymin ) / ( ymax - ymin ) ), 0, 1 );
		}
	} 

	/* Tipo de Objeto Invlido: nunca deve acontecer */
	return algVector( 0, 0, 0, 1 );	
}

int objGetMaterial( Object object )
{
	return object->material;
}

int objGetBumpFlag( Object object )
{
	return object->bump;
}

void objDestroy( Object object )
{
	free( object );
}


Vector objInterceptExit( Object object, Vector point, Vector d )
{
	switch( object->type )
	{
	case TYPE_SPHERE:
		{
			Sphere *s = (Sphere *)object->data;

			double a, b, c, delta, distance;
			//double distance = -1.0;
						
			Vector fromSphereToEye;

			fromSphereToEye = algSub( point, s->center );

			a = algDot( d, d );
			b = ( 2.0 * algDot( d, fromSphereToEye ) );
			c = ( algDot( fromSphereToEye, fromSphereToEye ) - ( s->radius * s->radius ) );

			delta = ( ( b * b ) - ( 4 * a * c ) );

			if( fabs( delta ) <= EPSILON )
			{
				distance = ( -b / (2 * a ) );
			}
			else if( delta > EPSILON )
			{
				double root = sqrt( delta );
				distance = MAX( ( ( -b + root ) / ( 2 * a ) ), ( ( -b - root ) / ( 2.0 * a ) )  );
			}

			return algAdd(point, algScale(distance, d));
		}

	case TYPE_TRIANGLE:
	case TYPE_BOX:
	    break;
	}
	return point;
}

Vector calculaNormais(Image bumpTga, Vector point)
{
	if(bumpTga != NULL){
	int width , height , x , y;
	
	
		Color colorAx, colorBx, colorC, colorAy, colorBy;
		Vector u, v ,c ,cAx,cBx,cAy,cBy;
	
		imageGetDimensions( bumpTga , &width, &height );
		x = ( (int)( point.x * ( width - 1 ) ) % width );
		y = ( (int)( point.y * ( height - 1 ) ) % height );

		colorC  = imageGetPixel(bumpTga, x, y);
		c = algVector( point.x, point.y, colorC.red*255., 1.0);

		colorAx = imageGetPixel(bumpTga, x - 1, y);
		cAx = algVector( x - 1.0, y, colorAx.red*255., 1.0);

		colorBx = imageGetPixel(bumpTga, x + 1, y);
		cBx = algVector( x + 1.0, y, colorBx.red*255., 1.0);

		colorAy = imageGetPixel(bumpTga, x, y - 1);
		cAy = algVector( x, y - 1.0, colorAy.red*255., 1.0);

		colorBy = imageGetPixel(bumpTga, x, y + 1);
		cBy = algVector( x, y + 1.0, colorBy.red*255., 1.0);

		u = algScale(0.5 ,algSub(cBx, cAx));
		v = algScale(0.5 ,algSub(cBy, cAy));

		return algScale(0.2,algCross(u, v));
	}
	return algVector(0.,0.,0.,1.);
}
 int maiorZ( Object object )
 {
	return  1;
 }

void objDraw( Object object )
{
    glPushAttrib(GL_ENABLE_BIT | GL_TRANSFORM_BIT | GL_CURRENT_BIT | GL_COLOR_BUFFER_BIT | GL_LIGHTING_BIT);

	if( object->type == TYPE_SPHERE )
	{
    	Sphere *sphere = (Sphere *)object->data;
    	//glMatrixMode(GL_MODELVIEW);
    	//glPushMatrix();
		glTranslated(sphere->center.x, sphere->center.y, sphere->center.z);
		glutSolidSphere(sphere->radius, 50, 50);
    	//glPopMatrix();
	}
	else if ( object->type == TYPE_BOX )
	{
    	Box *box = (Box *)object->data;
		Vector bottomLeft = box->bottomLeft;
		Vector topRight = box->topRight;
		double dx = topRight.x - bottomLeft.x;
		double dy = topRight.y - bottomLeft.y;
		double dz = topRight.z - bottomLeft.z;
		double cx = (topRight.x + bottomLeft.x) / 2;
		double cy = (topRight.y + bottomLeft.y) / 2;
		double cz = (topRight.z + bottomLeft.z) / 2;
		int temp;
		//glMatrixMode(GL_MODELVIEW);
		
		//glPushMatrix();
		glEnable(GL_TEXTURE_2D);
        glTranslated(cx, cy, cz);
		glScaled(dx/2, dy/2, dz/2);
		//glutSolidCube(1.0); 

        glBegin(GL_QUADS);
        // Front Face
        glNormal3f( 0.0, 0.0, 1.0);
        glTexCoord2f(0.0f, 0.0f); 
        glVertex3f(-1.0f, -1.0f,  1.0f);    // Bottom Left Of The Texture and Quad
        glTexCoord2f(1.0f, 0.0f); 
        glVertex3f( 1.0f, -1.0f,  1.0f);    // Bottom Right Of The Texture and Quad
        glTexCoord2f(1.0f, 1.0f); 
        glVertex3f( 1.0f,  1.0f,  1.0f);    // Top Right Of The Texture and Quad
        glTexCoord2f(0.0f, 1.0f); 
        glVertex3f(-1.0f,  1.0f,  1.0f);    // Top Left Of The Texture and Quad
        // Back Face
        glNormal3f( 0.0, 0.0,-1.0);
        glTexCoord2f(1.0f, 0.0f); 
        glVertex3f(-1.0f, -1.0f, -1.0f);    // Bottom Right Of The Texture and Quad
        glTexCoord2f(1.0f, 1.0f); 
        glVertex3f(-1.0f,  1.0f, -1.0f);    // Top Right Of The Texture and Quad
        glTexCoord2f(0.0f, 1.0f); 
        glVertex3f( 1.0f,  1.0f, -1.0f);    // Top Left Of The Texture and Quad
        glTexCoord2f(0.0f, 0.0f); 
        glVertex3f( 1.0f, -1.0f, -1.0f);    // Bottom Left Of The Texture and Quad
        // Top Face
        glNormal3f( 0.0, 1.0, 0.0);
        glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);    // Top Left Of The Texture and Quad
        glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,  1.0f,  1.0f);    // Bottom Left Of The Texture and Quad
        glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f,  1.0f,  1.0f);    // Bottom Right Of The Texture and Quad
        glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);    // Top Right Of The Texture and Quad
        // Bottom Face
        glNormal3f( 0.0,-1.0, 0.0);
        glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f, -1.0f, -1.0f);    // Top Right Of The Texture and Quad
        glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f, -1.0f, -1.0f);    // Top Left Of The Texture and Quad
        glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);    // Bottom Left Of The Texture and Quad
        glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);    // Bottom Right Of The Texture and Quad
        // Right face
        glNormal3f( 1.0, 0.0, 0.0);
        glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);    // Bottom Right Of The Texture and Quad
        glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);    // Top Right Of The Texture and Quad
        glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);    // Top Left Of The Texture and Quad
        glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);    // Bottom Left Of The Texture and Quad
        // Left Face
        glNormal3f(-1.0, 0.0, 0.0);
        glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);    // Bottom Left Of The Texture and Quad
        glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);    // Bottom Right Of The Texture and Quad
        glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);    // Top Right Of The Texture and Quad
        glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);    // Top Left Of The Texture and Quad
        glEnd();
		//glPopMatrix();
				
    }
	else if ( object->type == TYPE_TRIANGLE )
	{
	
    	Triangle *t = (Triangle *)object->data;
	    Vector vertice1 = t->v0;
	    Vector vertice2 = t->v1;
	    Vector vertice3 = t->v2;
        Vector v0ToV1 = algSub( t->v1, t->v0 );
        Vector v1ToV2 = algSub( t->v2, t->v1 );
        Vector normal = algUnit(algCross( v0ToV1, v1ToV2 ));

	    
	    glBegin(GL_TRIANGLES);
	    glNormal3f( normal.x,normal.y,normal.z);
	    glVertex3f(vertice1.x, vertice1.y, vertice1.z);
	    glVertex3f(vertice2.x, vertice2.y, vertice2.z);
	    glVertex3f(vertice3.x, vertice3.y, vertice3.z);
	    glEnd();
	    
	}
	
	glPopAttrib();
}