/**
 *	@file object.h Object: definio e operaes com primitivas.
 *		As primitivas suportadas atualmente so: esferas, tringulos e paraleleppedos.
 *
 *	@author
 *			- Maira Noronha
 *			- Thiago Bastos
 *			- Mauricio Carneiro
 *
 *	@date
 *			Criado em:			01 de Dezembro de 2002
 *			ltima Modificao:	22 de Janeiro de 2003
 *
 *	@version 2.0
 */

#ifndef _OBJECT_H_
#define _OBJECT_H_

#include "image.h"
#include "algebra.h"





/************************************************************************/
/* Tipos Exportados                                                     */
/************************************************************************/
/**
 *   Tipo objeto
 */
struct _Object
{
    /**
     *  Tipo do objeto.
     */
	int type;
	/**
     *  Material do objeto.
     */
	int material;
	/**
     *  Flag para indicar se deve aplicar Bump map caso haja textura.
     */
	int bump;
    /**
     *  Flag para indicar se deve aplicar Transparencia no Zbuffer.
     */
    int blend;
	/**
     *  Dados do objeto.
     */
	void *data;
};

/**
 *   Objeto esfera.
 */
typedef struct
{
	
    /**
	 *  Posio do centro da esfera na cena.
	 */
	Vector center;
	  
	/**
	 *  Raio da esfera.
	 */
	double radius;
}
Sphere;

/**
 *   Objeto tringulo.
 */
typedef struct
{	
	/**
	 *  Primeiro vrtice do tringulo.
	 */
	Vector v0;
	/**
	 *  Segundo vrtice do tringulo.
	 */
	Vector v1;
	/**
	 *  Terceiro vrtice do tringulo.
	 */
	Vector v2;
	
	Vector tex0;  /* coordenada de textura do verive 0 */
	Vector tex1;  /* coordenada de textura do verive 1 */
	Vector tex2;  /* coordenada de textura do verive 2 */
}
Triangle;

/**
 *   Objeto caixa.
 */
typedef struct
{	
	/**
	 *  Vrtice de baixo e  esquerda do paraleleppedo.
	 */
	Vector bottomLeft;	
	/**
	 *  Vrtice de cima e  direita do paraleleppedo.
	 */
	Vector topRight;
}
Box;

typedef struct _Object * Object;


/************************************************************************/
/* Funes Exportadas                                                   */
/************************************************************************/
/**
 *	Cria uma esfera.
 *
 *	@param material Id do material da esfera.
 *	@param center Posio do centro da esfera na cena.
 *	@param radius Raio da esfera.
 *
 *	@return Handle para o objeto criado.
 */
Object objCreateSphere( int material, const Vector center, double radius, int bumpFlag , int blendFlag );

/**
 *	Cria um tringulo.
 *
 *	@param material Id do material do tringulo.
 *	@param v0 Primeiro vrtice do tringulo.
 *	@param v1 Segundo vrtice do tringulo.
 *	@param v2 Terceiro vrtice do tringulo.
 *	@param tex0 Coordenadas de textura do primeiro vrtice do tringulo.
 *	@param tex1 Coordenadas de textura do segundo vrtice do tringulo.
 *	@param tex2 Coordenadas de textura do terceiro vrtice do tringulo.
 *
 *	@return Handle para o objeto criado.
 */
Object objCreateTriangle( int material, const Vector v0, const Vector v1, const Vector v2, 
						                const Vector tex0, const Vector tex1, const Vector tex2, int bumpFlag , int blendFlag);

/**
 *	Cria um paraleleppedo.
 *
 *	@param material Id do material do paraleleppedo.
 *	@param bottomLeft Vrtice de baixo e  esquerda do paraleleppedo.
 *	@param topRight Vrtice de cima e  direita do paraleleppedo.
 *
 *	@return Handle para o objeto criado.
 */
Object objCreateBox( int material, const Vector bottomLeft, const Vector topRight, int bumpFlag , int blendFlag);

/**
 *	Calcula a que distncia um raio intercepta um objeto.
 *
 *	@param object Handle para um objeto.
 *	@param eye Origem do raio.
 *	@param ray Direo do raio.
 *
 *	@return Distncia de eye at a superfcie do objeto no ponto onde ocorreu a
 *				interseo. Menor ou igual a zero se no houver interseo.
 */
double objIntercept( Object object, Vector eye, Vector ray );

/**
 *	Calcula o vetor normal a um objeto em um ponto.
 *
 *	@param object Handle para um objeto.
 *	@param point Ponto na superfcie do objeto onde a normal deve ser calculada.
 *
 *	@return Vetor unitrio, normal ao objeto, com origem em point.
 */
Vector objNormalAt( Object object, Vector point );

/**
 *	Calcula a coordenada de textura para um objeto em um ponto.
 *
 *	@param object Handle para um objeto.
 *	@param point Ponto na superfcie do objeto para onde uma coordenada de textura
 *					ser calculada.
 *
 *	@return Coordenada de textura para o objeto no ponto especificado.
 */
Vector objTextureCoordinateAt( Object object, Vector point );

/**
 *	Obtm o Material de um objeto.
 */
int objGetMaterial( Object object );

/**
 *	Destri um objeto criado com as funes objCreate*().
 */
void objDestroy( Object object );

Vector objInterceptExit( Object object, Vector point, Vector d );

Vector calculaNormais(Image bumpTga, Vector point);

int objGetBumpFlag( Object object );

void objDraw( Object object);

#endif