/**
 *	@file raytracing.c RayTracing: renderizao de cenas por raytracing.
 *
 *	@author 
 *			- Maira Noronha
 *			- Thiago Bastos
 *
 *	@date
 *			Criado em:			02 de Dezembro de 2002
 *			ltima Modificao:		4 de Junho de 2003
 *
 *	@version 2.0
 */

#include <math.h>
#include <float.h>
#include <string.h>
#include <stdlib.h>

#include "raytracing.h"
#include "color.h"
#include "algebra.h"


/************************************************************************/
/* Constantes Privadas                                                  */
/************************************************************************/

#ifndef EPSILON
#define EPSILON	1.0e-3
#endif

#define MAX( a, b ) ( ( a > b ) ? a : b )

/** Nmero mximo de recurses permitidas */
#define MAX_DEPTH	6


/************************************************************************/
/* Funes Privadas                                                     */
/************************************************************************/
/**
 *	Obtm uma cor atravs do traado de um raio dentro de uma cena.
 *
 *	@param scene Handle para a cena sendo renderizada.
 *	@param eye Posio do observador, origem do raio.
 *	@param ray Direo do raio.
 *	@param depth Para controle do nmero mximo de recurses. Funes clientes devem
 *					passar 0 (zero). A cada recurso depth  incrementado at, no mximo,
 *					MAX_DEPTH. Quando MAX_DEPTH  atingido, recurses so ignoradas.
 *
 *	@return Cor resultante do traado do raio.
 */
static Color shade( Scene scene, Vector eye, Vector ray, Object object, Vector point,
					Vector normal, int depth );

/**
 *	Encontra o primeiro objeto interceptado pelo raio originado na posio especificada.
 *
 *	@param scene Cena.
 *	@param eye Posio do Observador (origem).
 *	@param ray Raio sendo traado (direo).
 *	@param object Onde  retornado o objeto resultante. No pode ser NULL.
 *	@return Distncia entre 'eye' e a superfcie do objeto interceptado pelo raio.
 *			DBL_MAX se nenhum objeto  interceptado pelo raio, neste caso
 *				'object' no  modificado.
 */
static double getNearestObject( Scene scene, Vector eye, Vector ray, Object *object );

/**
 *	Checa se objetos em uma cena impedem a luz de alcanar um ponto.
 *
 *	@param scene Cena.
 *	@param point Ponto sendo testado.
 *	@param rayToLight Um raio (direo) indo de 'point' at 'lightLocation'.
 *	@param lightLocation Localizao da fonte de luz.
 *	@return Zero se nenhum objeto bloqueia a luz e no-zero caso contrrio.
 */
static int isInShadow( Scene scene, Vector point, Vector rayToLight, Vector lightLocation );

static Vector RaioCalota(Vector normal);


/************************************************************************/
/* Definio das Funes Exportadas                                     */
/************************************************************************/

Color rayTrace( Scene scene, Vector eye, Vector ray, int depth, int globalIlum )
{
 int i;
	Object object;
	double distance, total;
 Color pixel, pixelRaio;
 Color pixelIndireto = { 0.0, 0.0, 0.0 };

	Vector point , pontoTextura;
	Vector normal, raioCalota, pointRaioCalota;
	Material material;

	/* Calcula o primeiro objeto a ser atingido pelo raio */
	distance = getNearestObject( scene, eye, ray, &object );

	/* Se o raio no interceptou nenhum objeto... */
	if( distance == DBL_MAX )
		return sceGetBackgroundColor( scene, eye, ray );

	/* Calcula o ponto de interseo do raio com o objeto */
	point = algAdd( eye, algScale( distance, ray ) );

	/* Obtm o vetor normal ao objeto no ponto de interseo */
	normal =  objNormalAt( object, point );
 material = sceGetMaterial(scene,objGetMaterial(object));
	/* Bump map */
 if(temTextura( material ) && objGetBumpFlag( object ) )
	{
		pontoTextura = objTextureCoordinateAt( object, point );
		normal = calculaNormais( material->normal , pontoTextura );
	}
	
 pixel = shade( scene, eye, ray, object, point, algUnit(normal), depth );

 if(!globalIlum)
		return pixel;

	total = 0;
	for(i = 0; i < 500; i++)
 {
		raioCalota = RaioCalota(algUnit(normal));

		distance = getNearestObject( scene, point, raioCalota, &object );

		if( distance != DBL_MAX )
  {
			double cosTheta = algDot(normal, raioCalota);
			double pondTiro = sqrt(1 - cosTheta*cosTheta);
			total += pondTiro;

			pointRaioCalota = algAdd( point, algScale( distance, raioCalota ) );

			normal =  objNormalAt( object, pointRaioCalota );

			pixelRaio = shade( scene, eye, raioCalota, object, pointRaioCalota, algUnit(normal), 1, FALSE);
			pixelIndireto = colorContribute(pixelIndireto, colorScale(pondTiro, pixelRaio));
		}
	}

	if(total)
 {
		Color contribIndVersusDireta = colorMultiplication(pixel, pixelIndireto);
		return colorAddition(pixel, colorScale(1/((double)total), contribIndVersusDireta));
	}

	return pixel;
}

/************************************************************************/
/* Definio das Funes Privadas                                       */
/************************************************************************/

static Color shade( Scene scene, Vector eye, Vector ray, Object object, Vector point,
					Vector normal, int depth )
{
	int i;
	Light light;
	double k ;
	Vector L, r, v    ;
	Color colorEspecular, colorDifusa, rcolor;
	
   
	Material material = sceGetMaterial(scene,objGetMaterial(object));

	double reflectionFactor = matGetReflectionFactor( material );
	double specularExponent = matGetSpecularExponent( material );
	double refractedIndex   = matGetRefractionIndex( material );
	double opacity = matGetOpacity( material );

	Color ambient = sceGetAmbientLight( scene );
	Color diffuse = matGetDiffuse( material, objTextureCoordinateAt( object, point ) );	
	Color specular = matGetSpecular( material );
	
	/* Comea com a cor ambiente */
	Color color = colorMultiplication( diffuse, ambient );
	
	/*Completando a funcao*/
	for(i = 0; i < sceGetLightCount( scene ); i++)
	{
	    light = sceGetLight( scene, i );
	    
	    /*vetor unitrio na posicao de point para a posicao da luz*/
	    v = algUnit( algSub(eye, point));
        L = algUnit( algSub(lightGetPosition(light), point));
		normal = algUnit(normal);
		r = algReflect( L , normal ) ;
        
	    if(algDot(L, normal) > 0)    
	    {
	       if(! isInShadow( scene , point , L, lightGetPosition(light) ) )
	        {
	            /*componente difusa*/
				colorDifusa = colorScale( algDot( normal , L), colorMultiplication( lightGetColor(light), diffuse ) );
	            
	            /*componente especular*/
	            colorEspecular = colorScale( (pow(algDot(r, v), matGetSpecularExponent( material ))), colorMultiplication( lightGetColor(light) , specular ) ) ;
	            color = colorAddition(color, colorAddition( colorDifusa, colorEspecular ));
	        }
	    }
	}

	/*incluindo efeito de superfcies refletoras*/
	if(depth >= MAX_DEPTH) return color ;
	
	k = matGetReflectionFactor( material );
	
	if(k != 0)
	{
		/*raio na direcao de reflexao*/
		r = algUnit(algReflect(v, normal));

		/*Calcula a cor correspondente ao raio*/
		rcolor = rayTrace( scene, point, r, depth + 1, globalIlumination);
		color = colorAddition(color, colorScale(k, rcolor));
	}

	/*incluindo efeito de superfcies transparentes*/
	if(opacity < 1)
	{
	    Color refractedColor;
	    Vector refractedRay = algSnell(ray, normal, 1., refractedIndex);
	    point = objInterceptExit(object, point, refractedRay);
	    normal = algMinus(objNormalAt(object, point));
	    refractedRay = algSnell(refractedRay, normal, refractedIndex, 1.);
	    if(algNorm(refractedRay) > 1.e-4)
	    {
	        refractedColor = rayTrace(scene, point, refractedRay, depth + 1, globalIlumination);
	        color = colorAddition(color, colorScale((1 - opacity), refractedColor));
	    }
	}

	return color;
}

static double getNearestObject( Scene scene, Vector eye, Vector ray, Object *object )
{
	int i;
	int objectCount = sceGetObjectCount( scene );

	double closest = DBL_MAX;

	/* Para cada objeto na cena */
	for( i = 0; i < objectCount; ++i )
	{
		Object currentObject = sceGetObject( scene, i );
		double distance = objIntercept( currentObject, eye, ray );
		
		if( distance > 0.0 && distance < closest )
		{
			closest = distance;
			*object = currentObject;
		}
	}

	return closest;
}

static int isInShadow( Scene scene, Vector point, Vector rayToLight, Vector lightLocation )
{
	int i;
	int objectCount = sceGetObjectCount( scene );

	/* maxDistance = distncia de point at lightLocation */
	double maxDistance = algNorm( algSub( lightLocation, point ) );

	/* Para cada objeto na cena */
	for( i = 0; i < objectCount; ++i )
	{
		double distance = objIntercept( sceGetObject( scene, i ), point, rayToLight );
		
		if( distance > 0.1 && distance < maxDistance )
		{
			return 1;
		}
	}

	return 0;
}


Vector RaioCalota(Vector normal)
{
	double rand1 = rand() / ((double)RAND_MAX + 1);
	double rand2 = rand() / ((double)RAND_MAX + 1);

	double theta = 2.0 * MPI * rand2;
	Vector u, v, sum, scale, res;

	if(normal.x > 0.99 || normal.x < -0.99)
 {
		u.x = 0;
  u.y = 1;
  u.z = 0;
  u.w = 1;
 }
	else
 {
		u.x = 1;
  u.y = 0;
  u.z = 0;
  u.w = 1;
 }

	v = algCross(normal, u);
	u = algCross(v, normal);
	sum = algAdd(algScale(sqrt(rand1)*cos(theta), u), algScale(sqrt(rand1)*sin(theta), v));
	scale = algScale(sqrt(1.0 - rand1), normal);
	res = algAdd(sum, scale);
	
	return algUnit(res);
}
