/**
 *	@file scene.c Scene: definio e manuteno de cenas.
 *
 *	@author 
 *			- Maira Noronha
 *			- Thiago Bastos
 *
 *	@date
 *			Criado em:			02 de Dezembro de 2002
 *			ltima Modificao:	4 de Junho de 2003
 *
 *	@version 2.0
 */

#include "scene.h"
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/timeb.h>


/************************************************************************/
/* Definio das Funes Exportadas                                     */
/************************************************************************/
Color sceGetBackgroundColor( Scene scene, Vector eye, Vector ray )
{
	double planeDistance, divisor, distance, scaleU, scaleV;
	Vector farOrigin, farNormal, farU, farV, point, pointFromOrigin;

	if( scene->bgImage == NULL || scene->camera == NULL )
	{
		return scene->bgColor;
	}

	/* Obtm informaes sobre o far plane */
	camGetFarPlane( scene->camera, &farOrigin, &farNormal, &farU, &farV );

	/* Cos(alpha) entre a normal do plano e o raio */
	divisor = algDot( ray, farNormal );

	/* Se o raio se distancia ou  paralelo ao far plane */
	if( divisor > 0 || -divisor < EPSILON )
	{
		return scene->bgColor;
	}

	planeDistance = algDot( farOrigin, farNormal );

	distance = ( ( planeDistance - algDot( eye, farNormal ) ) / divisor );

	/* Se o raio se distancia do far plane */
	if( distance < 0 )
	{
		return scene->bgColor;
	}

	point = algAdd( eye, algScale( distance, ray ) );

	pointFromOrigin = algSub( point, farOrigin );

	divisor = algDot( farU, farU );
	scaleU = ( algDot( farU, pointFromOrigin ) / divisor );

	divisor = algDot( farV, farV );
	scaleV = ( algDot( farV, pointFromOrigin ) / divisor );

	/* Se o raio no intercepta o far plane (ou seja, a imagem de fundo)... */
	if( scaleU < 0 || scaleV < 0 || scaleU > 1 || scaleV > 1 )
	{
		return scene->bgColor;
	}
	
	return imageGetPixel( scene->bgImage,
							(int)( scaleU * camGetScreenWidth( scene->camera ) ),
							(int)( scaleV * camGetScreenHeight( scene->camera ) ) );
}

Color sceGetAmbientLight( Scene scene )
{
	return scene->ambientLight;
}

Camera sceGetCamera( Scene scene )
{
    return scene->camera;
}

int sceGetObjectCount( Scene scene )
{
	return scene->objectCount;
}

Object sceGetObject( Scene scene, int index )
{
	if( index < 0 || index >= scene->objectCount )
	{
		return NULL;
	}

	return scene->objects[index];
}

int sceGetLightCount( Scene scene )
{
	return scene->lightCount;
}

Light sceGetLight( Scene scene, int index )
{
	if( index < 0 || index >= scene->lightCount )
	{
		return NULL;
	}

	return scene->lights[index];
}

Scene sceLoad( const char *filename )
{
	FILE *file;
	char buffer[512];

	Scene scene;
	
	Color bgColor;
	Color ambientLight;
	char backgroundFileName[FILENAME_MAXLEN];

	/* Camera */
	Vector eye = algVector( 0,0,0,1 );
	Vector at = algVector( 0,0,0,1 );
	Vector up = algVector( 0,0,0,1 );
	double fovy;
	double nearp;
	double farp;
	int screenWidth;
	int screenHeight;

	/* Material */
	Color diffuse;
	Color specular;
	double specularExponent;
	double reflective;
	double refractive;
	double opacity;
	char textureFileName[FILENAME_MAXLEN];
	
	/* Lights & Objects */
	int i;
	double  bump ;
	int blendFlag;
	Color lightColor, dLightColor;
	int material;
	Vector pos1 = algVector( 0,0,0,1 );
	Vector pos2 = algVector( 0,0,0,1 );
	Vector pos3 = algVector( 0,0,0,1 );
	Vector tex1 = algVector( 0,0,0,1 );
	Vector tex2 = algVector( 0,0,0,1 );
	Vector tex3 = algVector( 0,0,0,1 );
	double radius;
	
	file = fopen( filename, "rt" );
	if( !file )
	{
		return NULL;
	}

	scene = (struct _Scene *)malloc( sizeof(struct _Scene) );
	if( !scene )
	{
		return NULL;
	}

	/* Default (undefined) values: */
	scene->camera = NULL;
	scene->bgImage = NULL;
	scene->objectCount = 0;
	scene->lightCount = 0;
	scene->materialCount = 0;
	
	while( fgets( buffer, sizeof(buffer), file ) ) 
	{
		if( sscanf( buffer, "RT %lf\n", &at ) == 1 )
		{
			/* Ignore File Version Information */
		}
		else if( sscanf( buffer, "CAMERA %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %d %d\n", &eye.x, &eye.y, &eye.z, &at.x, &at.y, &at.z, &up.x, &up.y, &up.z, &fovy, &nearp, &farp, &screenWidth, &screenHeight ) == 14) 
		{
			if( scene->camera )
			{
				camDestroy( scene->camera );
			}

			scene->camera = camCreate( eye, at, up, fovy, nearp, farp, screenWidth, screenHeight );
		} 
		else if( sscanf( buffer, "SCENE %lf %lf %lf %lf %lf %lf %s\n", &bgColor.red, &bgColor.green, &bgColor.blue, &ambientLight.red, &ambientLight.green, &ambientLight.blue, backgroundFileName ) == 7 ) 
		{
			bgColor = colorNormalize( bgColor );
			ambientLight = colorNormalize( ambientLight );

			scene->bgColor = bgColor;
			scene->ambientLight = ambientLight;

			if( scene->bgImage )
			{
				imageDestroy( scene->bgImage );
			}

			if( strcmp( backgroundFileName, "null") == 0 )
			{
				scene->bgImage = NULL;
			} 
			else 
			{
				scene->bgImage = imageLoad( backgroundFileName );
			}
		} 
		else if( sscanf( buffer, "MATERIAL %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %s\n", &diffuse.red, &diffuse.green, &diffuse.blue, &specular.red, &specular.green, &specular.blue, &specularExponent, &reflective, &refractive, &opacity, textureFileName ) == 11 ) 
		{
			Image image = NULL;
			TextureImage * texturaZbuffer = NULL;
			
			
			if( strcmp( textureFileName, "null") != 0 )
			{
				image = imageLoad( textureFileName );
				texturaZbuffer = (TextureImage *) malloc(sizeof(TextureImage));
				LoadTGA(texturaZbuffer,textureFileName);
			}

			if( scene->materialCount >= MAX_MATERIALS )
			{
				imageDestroy( image );
				fprintf( stderr, "sceLoad: Foi ultrapassado o limite de definicoes de materiais na cena. Ignorando." );
				continue;
			}

			diffuse = colorNormalize( diffuse );
			specular = colorNormalize( specular );

			scene->materials[scene->materialCount++] = matCreate( image,texturaZbuffer, diffuse, specular, specularExponent, reflective, refractive, opacity );
		} 
		//else if( sscanf( buffer, "LIGHT %lf %lf %lf %lf %lf %lf\n", &pos1.x, &pos1.y, &pos1.z, &lightColor.red, &lightColor.green, &lightColor.blue ) == 6 )
		//{
		//	if( scene->lightCount >= MAX_LIGHTS )
		//	{
		//		fprintf( stderr, "sceLoad: Foi ultrapassado o limite de definicoes de luzes na cena. Ignorando." );
		//		continue;
		//	}
  //          
  //          /* Dividindo a intensidade rgb da luz por 16 */
		//	lightColor  = colorNormalize( lightColor );
  //          dLightColor.red   = lightColor.red   / 16;
  //          dLightColor.green = lightColor.green / 16;
  //          dLightColor.blue  = lightColor.blue  / 16;
  //          
  //          /* Achando a posicao das novas luzes que deverao ser criadas */
  //          for(i = 0; i < 16; i++)
  //          {
  //              pos1.x += 10 * sin(i * 22.5);
  //              pos1.z += 10 * cos(i * 22.5);
  //              
  //              scene->lights[scene->lightCount++] = lightCreate( pos1, dLightColor );
  //          }  
		//} 
       else if( sscanf( buffer, "LIGHT %lf %lf %lf %lf %lf %lf\n", &pos1.x, &pos1.y, &pos1.z, &lightColor.red, &lightColor.green, &lightColor.blue ) == 6 )
		{
			if( scene->lightCount >= MAX_LIGHTS )
			{
				fprintf( stderr, "sceLoad: Foi ultrapassado o limite de definicoes de luzes na cena. Ignorando." );
				continue;
			}

			lightColor = colorNormalize( lightColor );

			scene->lights[scene->lightCount++] = lightCreate( pos1, lightColor );
		}
       else if( sscanf( buffer, "SPHERE %d %lf %lf %lf %lf %d\n", &material, &radius, &pos1.x,&pos1.y,&pos1.z,&blendFlag ) == 6 ) 
           {
           if( scene->objectCount >= MAX_OBJECTS )
               {
               fprintf( stderr, "sceLoad: Foi ultrapassado o limite de definicoes de objetos na cena. Ignorando." );
               continue;
               }

           scene->objects[scene->objectCount++] = objCreateSphere( material, pos1, radius, 0 ,blendFlag);
           } 
		else if( sscanf( buffer, "SPHERE %d %lf %lf %lf %lf\n", &material, &radius, &pos1.x,&pos1.y,&pos1.z ) == 5 ) 
		{
			if( scene->objectCount >= MAX_OBJECTS )
			{
				fprintf( stderr, "sceLoad: Foi ultrapassado o limite de definicoes de objetos na cena. Ignorando." );
				continue;
			}

			scene->objects[scene->objectCount++] = objCreateSphere( material, pos1, radius, 0 ,0);
		} 
        
		else if( sscanf( buffer, "TRIANGLE %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf\n", &material, &pos1.x, &pos1.y, &pos1.z, &pos2.x, &pos2.y, &pos2.z, &pos3.x, &pos3.y, &pos3.z, &tex1.x, &tex1.y, &tex2.x, &tex2.y, &tex3.x, &tex3.y) == 16 ) 
		{
			if( scene->objectCount >= MAX_OBJECTS )
			{
				fprintf( stderr, "sceLoad: Foi ultrapassado o limite de definicoes de objetos na cena. Ignorando." );
				continue;
			}
			
			scene->objects[scene->objectCount++] = objCreateTriangle( material, pos1, pos2, pos3, tex1, tex2, tex3 , 0 ,0);
		}
		else if( sscanf( buffer, "BOX %d %lf %lf %lf %lf %lf %lf %lf \n", &material, &pos1.x, &pos1.y, &pos1.z, &pos2.x, &pos2.y, &pos2.z,&bump ) == 8 ) 
		{
			if( scene->objectCount >= MAX_OBJECTS )
			{
				fprintf( stderr, "sceLoad: Foi ultrapassado o limite de definicoes de objetos na cena. Ignorando." );
				continue;
			}

			scene->objects[scene->objectCount++] = objCreateBox( material, pos1, pos2, bump ,0);
		} 
	  	else if( sscanf( buffer, "BOX %d %lf %lf %lf %lf %lf %lf %lf \n", &material, &pos1.x, &pos1.y, &pos1.z, &pos2.x, &pos2.y, &pos2.z,&bump ) == 7 ) 
		{
			if( scene->objectCount >= MAX_OBJECTS )
			{
				fprintf( stderr, "sceLoad: Foi ultrapassado o limite de definicoes de objetos na cena. Ignorando." );
				continue;
			}

			scene->objects[scene->objectCount++] = objCreateBox( material, pos1, pos2,0 ,0);
		}
		
		else
		{			
			printf( "sceLoad: Ignorando comando:\n %s\n", buffer );
		}
	}

	/* Adjust background image to screen size */
	if( scene->camera && scene->bgImage )
	{
		scene->bgImage = imageResize( scene->bgImage, camGetScreenWidth( scene->camera ),
			camGetScreenHeight( scene->camera ) );
	}

	fclose( file );

	return scene;
}


int sceGetMaterialCount( Scene scene )
{
	return scene->materialCount;
}

Material sceGetMaterial( Scene scene, int index )
{
	return scene->materials[index];
}

void sceDestroy( Scene scene )
{
	int i;

	camDestroy( scene->camera );
	imageDestroy( scene->bgImage );

	for( i = 0; i < scene->objectCount; ++i )
	{
		objDestroy( scene->objects[i] );
	}

	for( i = 0; i < scene->materialCount; ++i )
	{
		matDestroy( scene->materials[i] );
	}
	
	free( scene );
}
