/**
 *	@file scene.h Scene: definio e manuteno de cenas.
 *
 *	@author 
 *			- Maira Noronha
 *			- Thiago Bastos
 *
 *	@date
 *			Criado em:			02 de Dezembro de 2002
 *			ltima Modificao:	4 de Junho de 2003
 *
 *	@version 2.0
 */

#ifndef _SCENE_H_
#define _SCENE_H_

#include "image.h"
#include "light.h"
#include "camera.h"
#include "object.h"
#include "material.h"


/************************************************************************/
/* Constantes Exportadas                                                */
/************************************************************************/
#define MAX_MATERIALS	64
#define MAX_OBJECTS		128
#define MAX_LIGHTS		100
#define FILENAME_MAXLEN	64

#ifndef EPSILON
#define EPSILON	1.0e-10
#endif


/************************************************************************/
/* Tipos Exportados                                                     */
/************************************************************************/
/**
 *   Cena com a camera, os objetos, as luzes e a imagem/cor de fundo.
 */
struct _Scene
{
	/**
     *  Camera da cena
     */
	Camera camera;

    /**
     *  Cor de fundo da cena
     */
	Color bgColor;
	/**
     *  Intensidade rgb da luz ambiente da cena
     */
	Color ambientLight;
	/**
     *  Imagem de fundo da cena
     */
	Image bgImage;

	/**
     *  Nmero de materiais existentes na cena.
     */
	int materialCount;
	/**
     *  Vetor com os materiais existentes na cena.
     */
	Material materials[MAX_MATERIALS];

	/**
     *  Nmero de objetos existentes na cena.
     */
	int objectCount;
	/**
     *  Vetor com os objetos existentes na cena.
     */
	Object objects[MAX_OBJECTS];

	/**
     *  Nmero de fontes de luz existentes na cena.
     */
	int lightCount;
	/**
     *  Vetor com as fontes de luz existentes na cena.
     */
	Light lights[MAX_LIGHTS];
};

typedef struct _Scene * Scene;


/************************************************************************/
/* Funes Exportadas                                                   */
/************************************************************************/
/**
 *	Obtm uma cor de fundo traando um raio dentro de uma cena.
 *	 necessrio que uma cmera tenha sido definida para a cena.
 *	Se o raio atingir o far plane e houver uma imagem de fundo, a cor  retirada dessa
 *	imagem. Seno, a cor de fundo padro  retornada.
 *
 *	@param scene Handle para uma cena.
 *	@param eye Origem do raio.
 *	@param ray Direo do raio.
 *
 *	@return Cor de fundo para o raio fornecido na cena fornecida.
 */
Color sceGetBackgroundColor( Scene scene, Vector eye, Vector ray );

/**
 *	Obtm a luz ambiente de uma cena.
 */
Color sceGetAmbientLight( Scene scene );

/**
 *	Obtm a cmera de uma cena.
 */
Camera sceGetCamera( Scene scene );

/**
 *	Obtm o nmero de objetos existentes em uma cena.
 *
 *	@param scene Handle para uma cena.
 *
 *	@return Nmero de objetos na cena.
 */
int sceGetObjectCount( Scene scene );

/**
 *	Obtm um objeto de uma cena.
 *
 *	@param scene Handle para uma cena.
 *	@param index ndice do objeto a ser obtido (de 0 a objectCount - 1).
 *
 *	@return Objeto no ndice especificado (NULL se o ndice for invlido).
 */
Object sceGetObject( Scene scene, int index );

/**
 *	Obtm o nmero de fontes de luz existentes em uma cena.
 *
 *	@param scene Handle para uma cena.
 *	
 *	@return Nmero de luzes na cena.
 */
int sceGetLightCount( Scene scene );

/**
 *	Obtm uma fonte de luz de uma cena.
 *
 *	@param scene Handle para uma cena.
 *	@param index ndice da fonte de luz a ser obtida (de 0 a lightCount - 1).
 *
 *	@return A fonte de luz no ndice especificado (NULL se o ndice for invlido).
 */
Light sceGetLight( Scene scene, int index );

/**
 *	L uma cena a partir de um arquivo em formato rt4.
 *
 *	@param filename nome do arquivo que contm a cena.
 *
 *	@return Cena criada (NULL se o arquivo for invlido).
 */
Scene sceLoad( const char *filename );

/**
 *	Obtm o nmero de materiais de uma cena.
 */
int sceGetMaterialCount( Scene scene );

/**
 *	Obtm um material de uma cena.
 *
 *	@param scene Handle para uma cena.
 *	@param index ndice do material (de 0 a materialCount - 1).
 *
 *	@return O material no ndice especificado (NULL se o ndice for invlido).
 */
Material sceGetMaterial( Scene scene, int index );

/**
 *	Destri uma cena.
 */
void sceDestroy( Scene scene );

#endif
